/*******************************************

 PMACH.CPP - The KINT Interpreter
	Copyright 2002 Everett Kaser
	All rights reserved.

	Do not copy or redistribute without the
	author's explicit written permission.
 *******************************************/
#include <windows.h>
#include <io.h>
#include <fcntl.h>
#include <stdio.h>

#include	"main.h"
#include	"pmach.h"
#include	"sysfuncs.h"

//************************************************************
//* PORTING NOTES
//* -------------
//* 1) In this code, a BYTE is 8 bits, a WORD is 16 bits,
//*		a DWORD is 32 bits, all unsigned.  A 'char' is 8 bits,
//*		a 'short' is 16 bits, a long is 32 bits, all signed.
//* 2) Multi-byte values in the KODEs are stored in little-endian
//*		order (what's used on current INTEL processors, the primary
//*		current target platform.  If you're porting this to a
//*		platform that uses a big-endian processor, you'll need to
//*		VERY carefully modify the multi-byte operations to grab
//*		the bytes in the correct order. See the GETREG/SETREG and
//*		GETMEM/SETMEM macros in pmach.h.
//* 3) This implementation knows that memory pointers are DWORDS
//*		and uses that fact to place them directly into "registers"
//*		R0, R1, and R2.  (The .KE code only saves and loads these
//*		values, NEVER does arithmetic on them.)  If this gets ported
//*		to a machine that uses (for example) 8-byte pointers, then
//*		those pointers would need to be saved in a table or structure,
//*		and a HANDLE (or index) would be placed in R0, R1, and R2
//*		in place of the actual memory pointer.  Then, the actual
//*		memory pointer would have to be fetched anytime it was needed
//*		by the execution of the .KE kodes.  Places where this is
//*		a concern, I've tried to mark with "WARN: PTR4".
//*		The same applies to MS Windows HANDLES (such as HDCs, etc).
//* 4) Wherever possible, I've tried to use as simple, lowest-level
//*		C-library functions as possible in order to keep the code
//*		as portable as possible, so that folks doing ports don't
//*		have to modify any more code than necessary.
//************************************************************

//************************************************************

extern char		PgmPath[512], TmpPath[512];
extern HWND		hWndMain;
extern HDC		kDC;

//************************************************************

BYTE	mainRegs[256]; // 4*32 + 2*48 + 1*32 = 256 (32 DWORDS, 48 WORDS, 32 BYTES)
BYTE	*mainKode, *mainKodeEnd;
BYTE	gateEvent=0, killpgm=0;
DWORD	result;	// MSbit, EQ, LSbit
BYTE	signedmath;
DWORD	KEmemlen, StackBase;

//************************************************************
//* Read in .KE, allocate memory for it, initialize ptrs
//*		Return TRUE if successful, FALSE if not.
//*
BOOL	LoadKEFile(char *pnameKE) {

	int		fh, readcnt;
	BYTE	*pB;
	DWORD	remains, zerosize;

// Open .KE file
	strcpy(TmpPath, PgmPath);
	strcat(TmpPath, pnameKE);
	if( -1==(fh=_open(TmpPath, _O_BINARY | _O_RDONLY)) ) {
		MessageBox(NULL, ".KE file not found", "Oops", MB_OK);
		return FALSE;
	}
// Read .KE header
	if( 20!=(readcnt=_read(fh, TmpPath, 20)) ) {
		MessageBox(NULL, "Error reading .KE file", "Oops", MB_OK);
		_close(fh);
		return FALSE;
	}
	if( strcmp(TmpPath, "KASMB") ) {
		MessageBox(NULL, "Not a .KE file", "Oops", MB_OK);
		_close(fh);
		return FALSE;
	}
	if( ((BYTE)TmpPath[KHDR_MAJOR_INTERP_REV] & 0x80)
	 || ((BYTE)TmpPath[KHDR_MAJOR_INTERP_REV] > INTERP_MAJOR_REV)
	 || ((BYTE)TmpPath[KHDR_MAJOR_INTERP_REV]==INTERP_MAJOR_REV
		 && (BYTE)TmpPath[KHDR_MINOR_INTERP_REV]>INTERP_MINOR_REV) ) {
		MessageBox(NULL, "This .KE file requires a newer KINT.EXE", "Oops", MB_OK);
		_close(fh);
		return FALSE;
	}

// Get size of initialized code/data in file
	remains = GETMEMD(TmpPath+KHDR_CODE_SIZE);
	zerosize = GETMEMD(TmpPath+KHDR_ZERODATA_SIZE);
// Save pointer to base of stack
	StackBase = remains + zerosize;
// Get size of memory required (init code/data size + uninit data size + stack size)
	KEmemlen = remains + zerosize + GETMEMD(TmpPath+KHDR_STACK_SIZE);
// Allocate .KE memory
	if( NULL==(mainKode=(BYTE *)malloc(KEmemlen)) ) {
		MessageBox(NULL, "Error allocating .KE memory", "Oops", MB_OK);
		_close(fh);
		return FALSE;
	}
// Copy first 20 bytes that we already read, and setup ptr for rest of reading
	memcpy(mainKode, TmpPath, 20);
	pB = mainKode + 20;
	mainKodeEnd = mainKode+KEmemlen;
	remains -= 20;
// Read rest of .KE file into memory
// Do it in 16K chunks in case some C library doesn't support large reads
	while( remains && readcnt ) {
		if( -1==(readcnt=_read(fh, pB, min(16384, remains))) ) {
			_close(fh);
			return FALSE;
		}
		pB += readcnt;
		remains -= (DWORD)readcnt;
	}
	_close(fh);
// Initialize uninitialized area to 0.
// Make sure that your memset takes at least a 4-byte value for "length".
// Otherwise, you'll need to modify this next line to properly zero-out
// all of the ZERODATA area, which may be longer than 64K bytes.
	memset(mainKode+GETMEMD(TmpPath+KHDR_CODE_SIZE), 0, zerosize);
// Init mainRegs to 0
	memset(mainRegs, 0, sizeof(mainRegs));

	return TRUE;
}

//************************************************
//* The K interpreter
//*		Event Kode must put return value into R23 before HALTing
//*
//* NOTE: "result" contains the KCPU "flags" for equality,
//*		least-significant bit, and most-significant bit.
//*		The "result" flags are set by these instructions:
//*			SHL SHR ROL ROR SUB B2I "RET 0" and "RET 1"
//*			AND XOR CMP INC DEC NOT NEG TST
//*		When the operation is a WORD or BYTE, the result
//*		is first cast to a SHORT or a CHAR before being
//*		cast to a DWORD in order to make sure that the
//*		most-significant bit gets propagated correctly.
//*
//*		The instructions that are affected by the 'signedmath' flag are:
//*			SHR, DIV, MOD, SUB, CMP

WORD ExecKE(BYTE *Kregs, BYTE *pKode) {

	WORD	*PC, opcode;
	BYTE	*pMem;
	DWORD	dw, dw2;
	WORD	w, w2, dr;
	BYTE	b;

	PC = (WORD *)(pKode + GETREGD(REG_3));	// get local ptr to PC code

	while( TRUE ) {
		opcode = GETMEMW(PC++);
		switch( (opcode>>8) & 0x00FF ) {
		 case 0x00:	// unused
		 case 0x01:	// unused
		 case 0x02:	// unused
			break;
		 case 0x03:	// SHL b,b
			dr = opcode & 0x00FF;
			result = (DWORD)(SETREGB(dr, GETREGB(dr)<<(BYTE)(GETREGB(GETMEMW(PC++)) & 0x0007)));
			result |= (result<<24);
			break;
		 case 0x04:	// SHL w,b
			dr = opcode & 0x00FF;
			result = (DWORD)(SETREGW(dr, GETREGW(dr)<<(BYTE)(GETREGB(GETMEMW(PC++)) & 0x000F)));
			result |= (result<<16);
			break;
		 case 0x05:	// SHL d,b
			dr = opcode & 0x00FF;
			result = SETREGD(dr, GETREGD(dr)<<(BYTE)(GETREGB(GETMEMW(PC++)) & 0x001F));
			break;
		 case 0x06:	// SHR b,b
			dr = opcode & 0x00FF;
			if( signedmath ) result = (DWORD)(char)(SETREGB(dr, (char)(GETREGB(dr))>>(BYTE)(GETREGB(GETMEMW(PC++)) & 0x07)));
			else result = (DWORD)(char)(SETREGB(dr, GETREGB(dr)>>(BYTE)(GETREGB(GETMEMW(PC++)) & 0x07)));
			result |= (result<<24);
			break;
		 case 0x07:	// SHR w,b
			dr = opcode & 0x00FF;
			if( signedmath ) result = (DWORD)(short)(SETREGW(dr, (short)(GETREGW(dr))>>(BYTE)(GETREGB(GETMEMW(PC++)) & 0x000F)));
			else result = (DWORD)(short)(SETREGW(dr, GETREGW(dr)>>(BYTE)(GETREGB(GETMEMW(PC++)) & 0x000F)));
			result |= (result<<16);
			break;
		 case 0x08:	// SHR d,b
			dr = opcode & 0x00FF;
			if( signedmath ) result = SETREGD(dr, (long)(GETREGD(dr))>>(BYTE)(GETREGB(GETMEMW(PC++)) & 0x001F));
			else result = SETREGD(dr, GETREGD(dr)>>(BYTE)(GETREGB(GETMEMW(PC++)) & 0x001F));
			break;
		 case 0x09: // ROL b,b
			dr = opcode & 0x00FF;
			w = (WORD)(GETREGB(GETMEMW(PC++)) & 0x07);
			b = GETREGB(dr);
			while( w-- ) {
				if( b & 0x80 ) b = (b << 1) | 1;
				else b <<= 1;
			}
			result = (DWORD)(char)SETREGB(dr, b);
			result |= (result<<24);
			break;
		 case 0x0A:	// ROL w,b
			dr = opcode & 0x00FF;
			b = GETREGB(GETMEMW(PC++)) & 0x1F;
			w = GETREGW(dr);
			while( b-- ) {
				if( w & 0x8000 ) w = (w<<1) | 1;
				else w <<= 1;
			}
			result = SETREGW(dr, w);
			break;
		 case 0x0B:	// ROL d,b
			dr = opcode & 0x00FF;
			dw = GETREGD(dr);
			w = GETREGB(GETMEMW(PC++)) & 0x1F;
			while( w-- ) {
				dw2 = dw;
				dw <<= 1;
				if( dw2 & 0x80000000 ) dw |= 1;
			}
			result = SETREGD(dr, dw);
			break;
		 case 0x0C:	// ROR b,b
			dr = opcode & 0x00FF;
			w = (WORD)(GETREGB(GETMEMW(PC++)) & 0x07);
			b = GETREGB(dr);
			while( w-- ) {
				if( b & 1 ) b = (b >> 1) | 0x80;
				else b >>= 1;
			}
			result = (DWORD)(char)SETREGB(dr, b);
			result |= (result<<24);
			break;
		 case 0x0D:	// ROR w,b
			dr = opcode & 0x00FF;
			b = GETREGB(GETMEMW(PC++)) & 0x001F;
			w = GETREGW(dr);
			while( b-- ) {
				if( w & 1 ) w = (w >> 1) | 0x8000;
				else w >>= 1;
			}
			result = (DWORD)(short)SETREGW(dr, w);
			result |= (result<<16);
			break;
		 case 0x0E:	// ROR d,b
			dr = opcode & 0x00FF;
			dw = GETREGD(dr);
			w = GETREGB(GETMEMW(PC++)) & 0x1F;
			while( w-- ) {
				dw2 = dw;
				dw >>= 1;
				if( dw2 & 1 ) dw |= 0x80000000;
			}
			result = SETREGD(dr, dw);
			break;
		 case 0x0F:	// ADD b,b
			dr = opcode & 0x00FF;
			SETREGB(dr, GETREGB(dr)+GETREGB(GETMEMW(PC++)));
			break;
		 case 0x10:	// ADD w,w
			dr = opcode & 0x00FF;
			SETREGW(dr, GETREGW(dr)+GETREGW(GETMEMW(PC++)));
			break;
		 case 0x11:	// ADD d,d
			dr = opcode & 0x00FF;
			SETREGD(dr, GETREGD(dr)+GETREGD(GETMEMW(PC++)));
			break;
		 case 0x12:	// SUB b,b
			dr = opcode & 0x00FF;
			if( signedmath ) result = (DWORD)((long)(((long)((char)GETREGB(dr)))-((long)((char)GETREGB(GETMEMW(PC++))))));
			else result = (DWORD)GETREGB(dr)-(DWORD)GETREGB(GETMEMW(PC++));
			SETREGB(dr, (BYTE)result);
			break;
		 case 0x13:	// SUB w,w
			dr = opcode & 0x00FF;
			if( signedmath ) result = (DWORD)((long)(((long)((short)GETREGW(dr)))-((long)((short)GETREGW(GETMEMW(PC++))))));
			else result = (DWORD)GETREGW(dr)-(DWORD)GETREGW(GETMEMW(PC++));
			SETREGW(dr, (WORD)result);
			break;
		 case 0x14:	// SUB d,d
			dr = opcode & 0x00FF;
			if( signedmath ) result = SETREGD(dr, (long)(GETREGD(dr)) - (long)GETREGD(GETMEMW(PC++)));
			else result = SETREGD(dr, GETREGD(dr) - GETREGD(GETMEMW(PC++)));
			break;
		 case 0x15:	// MUL b,b
			dr = opcode & 0x00FF;
			SETREGB(dr, GETREGB(dr)*GETREGB(GETMEMW(PC++)));
			break;
		 case 0x16:	// MUL w,w
			dr = opcode & 0x00FF;
			SETREGW(dr, GETREGW(dr)*GETREGW(GETMEMW(PC++)));
			break;
		 case 0x17:	// MUL d,d
			dr = opcode & 0x00FF;
			SETREGD(dr, GETREGD(dr)*GETREGD(GETMEMW(PC++)));
			break;
		 case 0x18:	// DIV b,b
			dr = opcode & 0x00FF;
			if( signedmath ) SETREGB(dr, (char)(GETREGB(dr))/(char)GETREGB(GETMEMW(PC++)));
			else SETREGB(dr, GETREGB(dr)/GETREGB(GETMEMW(PC++)));
			break;
		 case 0x19:	// DIV w,w
			dr = opcode & 0x00FF;
			if( signedmath ) SETREGW(dr, (short)(GETREGW(dr))/(short)GETREGW(GETMEMW(PC++)));
			else SETREGW(dr, GETREGW(dr)/GETREGW(GETMEMW(PC++)));
			break;
		 case 0x1A:	// DIV d,d
			dr = opcode & 0x00FF;
			if( signedmath ) SETREGD(dr, (long)(GETREGD(dr))/(long)(GETREGD(GETMEMW(PC++))));
			else SETREGD(dr, GETREGD(dr)/GETREGD(GETMEMW(PC++)));
			break;
		 case 0x1B:	// MOD b,b
			dr = opcode & 0x00FF;
			if( signedmath ) SETREGB(dr, (char)(GETREGB(dr))%(char)GETREGB(GETMEMW(PC++)));
			else SETREGB(dr, GETREGB(dr)%GETREGB(GETMEMW(PC++)));
			break;
		 case 0x1C:	// MOD w,w
			dr = opcode & 0x00FF;
			if( signedmath ) SETREGW(dr, (short)(GETREGW(dr))%(short)GETREGW(GETMEMW(PC++)));
			else SETREGW(dr, GETREGW(dr)%GETREGW(GETMEMW(PC++)));
			break;
		 case 0x1D:	// MOD d,d
			dr = opcode & 0x00FF;
			if( signedmath ) SETREGD(dr, (long)(GETREGD(dr))%(long)(GETREGD(GETMEMW(PC++))));
			else SETREGD(dr, GETREGD(dr)%GETREGD(GETMEMW(PC++)));
			break;
		 case 0x1E:	// OR b,b
			dr = opcode & 0x00FF;
			SETREGB(dr, GETREGB(dr) | GETREGB(GETMEMW(PC++)));
			break;
		 case 0x1F:	// OR w,w
			dr = opcode & 0x00FF;
			SETREGW(dr, GETREGW(dr) | GETREGW(GETMEMW(PC++)));
			break;
		 case 0x20:	// OR d,d
			dr = opcode & 0x00FF;
			SETREGD(dr, GETREGD(dr) | GETREGD(GETMEMW(PC++)));
			break;
		 case 0x21:	// AND b,b
			dr = opcode & 0x00FF;
			result = (DWORD)((long)(char)GETREGB(dr) & (long)(char)GETREGB(GETMEMW(PC++)));
			SETREGB(dr, (BYTE)result);
			break;
		 case 0x22:	// AND w,w
			dr = opcode & 0x00FF;
			result = (DWORD)((long)(short)GETREGW(dr) & (long)(short)GETREGW(GETMEMW(PC++)));
			SETREGW(dr, (WORD)result);
			break;
		 case 0x23:	// AND d,d
			dr = opcode & 0x00FF;
			result = SETREGD(dr, GETREGD(dr) & GETREGD(GETMEMW(PC++)));
			break;
		 case 0x24:	// XOR b,b
			dr = opcode & 0x00FF;
			result = (DWORD)(char)SETREGB(dr, GETREGB(dr) ^ GETREGB(GETMEMW(PC++)));
			result |= (result<<24);
			break;
		 case 0x25:	// XOR w,w
			dr = opcode & 0x00FF;
			result = (DWORD)(short)SETREGW(dr, GETREGW(dr) ^ GETREGW(GETMEMW(PC++)));
			result |= (result<<16);
			break;
		 case 0x26:	// XOR d,d
			dr = opcode & 0x00FF;
			result = SETREGD(dr, GETREGD(dr) ^ GETREGD(GETMEMW(PC++)));
			break;
		 case 0x27:	// CMP b,b
			if( signedmath ) result = (DWORD)((long)(((long)(char)GETREGB(opcode & 0x00FF))-((long)(char)(GETREGB(GETMEMW(PC++))))));
			else result = (DWORD)((DWORD)((long)((short)((char)((BYTE)GETREGB(opcode & 0x00FF)))))-(DWORD)((long)((short)((char)((BYTE)GETREGB(GETMEMW(PC++)))))));
			break;
		 case 0x28:	// CMP w,w
			if( signedmath ) result = (DWORD)((long)(((long)(short)GETREGW(opcode & 0x00FF))-((long)(short)(GETREGW(GETMEMW(PC++))))));
			else result = (DWORD)GETREGW(opcode & 0x00FF)-(DWORD)GETREGW(GETMEMW(PC++));
			break;
		 case 0x29:	// CMP d,d
			if( signedmath ) result = (long)(GETREGD(opcode & 0x00FF)) - (long)(GETREGD(GETMEMW(PC++)));
			else result = (DWORD)GETREGD(opcode & 0x00FF) - (DWORD)GETREGD(GETMEMW(PC++));
			break;
		 case 0x2A:	// TST b,b
			result = (DWORD)(char)(GETREGB(opcode & 0x00FF) & GETREGB(GETMEMW(PC++)));
			result |= (result<<24);
			break;
		 case 0x2B:	// TST w,w
			result = (DWORD)(short)(GETREGW(opcode & 0x00FF) & GETREGW(GETMEMW(PC++)));
			result |= (result<<16);
			break;
		 case 0x2C:	// TST d,d
			result = GETREGD(opcode & 0x00FF) & GETREGD(GETMEMW(PC++));
			break;
		 case 0x2D:	// XCH b,b
			dr = opcode & 0x00FF;
			b = GETREGB(dr);
			w = GETMEMW(PC++);
			SETREGB(dr, GETREGB(w));
			SETREGB(w, b);
			break;
		 case 0x2E:	// XCH w,w
			dr = opcode & 0x00FF;
			w2 = GETREGW(dr);
			w = GETMEMW(PC++);
			SETREGW(dr, GETREGW(w));
			SETREGW(w, w2);
			break;
		 case 0x2F:	// XCH d,d
			dr = opcode & 0x00FF;
			dw = GETREGD(dr);
			w = GETMEMW(PC++);
			SETREGD(dr, GETREGD(w));
			SETREGD(w, dw);
			break;
		 case 0x30:	// LDR b,b
			SETREGB(opcode & 0x00FF, GETREGB(GETMEMW(PC++)));
			break;
		 case 0x31:	// LDR w,w
			SETREGW(opcode & 0x00FF, GETREGW(GETMEMW(PC++)));
			break;
		 case 0x32:	// LDR d,d
			SETREGD(opcode & 0x00FF, GETREGD(GETMEMW(PC++)));
			break;
		 case 0x33:	// LDE d,b
			SETREGD(opcode & 0x00FF, (DWORD)(BYTE)GETREGB(GETMEMW(PC++)));
			break;
		 case 0x34:	// LDE d,w
			SETREGD(opcode & 0x00FF, (DWORD)(WORD)GETREGW(GETMEMW(PC++)));
			break;
		 case 0x35:	// LDE w,b
			SETREGW(opcode & 0x00FF, (WORD)GETREGB(GETMEMW(PC++)));
			break;
		 case 0x36:	// ABIT w,b
			dr = opcode & 0x00FF;
			if( dr<=REG_96 ) {
				w2 = GETMEMW(PC++);
				w  = GETREGW(dr);
				for(; w; w>>=1, ++w2) if( w & 1 ) SETREGB(w2, 1+GETREGB(w2));
			}
			break;
		 case 0x37:	// ABIT b,b
			w2 = GETMEMW(PC++);
			b  = GETREGB(opcode & 0x00FF);
			for(; b; b>>=1, ++w2) if( b & 1 ) SETREGB(w2, 1+GETREGB(w2));
			break;
		 case 0x38:	// LD0 b,d
			SETREGB(opcode & 0x00FF, GETMEMB(pKode + GETREGD(GETMEMW(PC++))));
			break;
		 case 0x39:	// LD0 w,d
			SETREGW(opcode & 0x00FF, GETMEMW(pKode + GETREGD(GETMEMW(PC++))));
			break;
		 case 0x3A:	// LD0 d,d
			SETREGD(opcode & 0x00FF, GETMEMD(pKode + GETREGD(GETMEMW(PC++))));
			break;
		 case 0x3B:	// LD1 b,d
			SETREGB(opcode & 0x00FF, GETMEMB(GETREGD(REG_1) + GETREGD(GETMEMW(PC++))));
			break;
		 case 0x3C:	// LD1 w,d
			SETREGW(opcode & 0x00FF, GETMEMW(GETREGD(REG_1) + GETREGD(GETMEMW(PC++))));
			break;
		 case 0x3D:	// LD1 d,d
			SETREGD(opcode & 0x00FF, GETMEMD(GETREGD(REG_1) + GETREGD(GETMEMW(PC++))));
			break;
		 case 0x3E:	// LD2 b,d
			SETREGB(opcode & 0x00FF, GETMEMB(GETREGD(REG_2) + GETREGD(GETMEMW(PC++))));
			break;
		 case 0x3F:	// LD2 w,d
			SETREGW(opcode & 0x00FF, GETMEMW(GETREGD(REG_2) + GETREGD(GETMEMW(PC++))));
			break;
		 case 0x40:	// LD2 d,d
			SETREGD(opcode & 0x00FF, GETMEMD(GETREGD(REG_2) + GETREGD(GETMEMW(PC++))));
			break;
		 case 0x41:	// LDS b,d
			SETREGB(opcode & 0x00FF, GETMEMB(pKode + GETREGD(REG_4) + GETREGD(GETMEMW(PC++))));
			break;
		 case 0x42:	// LDS w,d
			SETREGW(opcode & 0x00FF, GETMEMW(pKode + GETREGD(REG_4) + GETREGD(GETMEMW(PC++))));
			break;
		 case 0x43:	// LDS d,d
			SETREGD(opcode & 0x00FF, GETMEMD(pKode + GETREGD(REG_4) + GETREGD(GETMEMW(PC++))));
			break;
		 case 0x44:	// ST0 b,d
			SETMEMB(pKode + GETREGD(GETMEMW(PC++)), GETREGB(opcode & 0x00FF));
			break;
		 case 0x45:	// ST0 w,d
			SETMEMW(pKode + GETREGD(GETMEMW(PC++)), GETREGW(opcode & 0x00FF));
			break;
		 case 0x46:	// ST0 d,d
			SETMEMD(pKode + GETREGD(GETMEMW(PC++)), GETREGD(opcode & 0x00FF));
			break;
		 case 0x47:	// ST1 b,d
			SETMEMB(GETREGD(REG_1) + GETREGD(GETMEMW(PC++)), GETREGB(opcode & 0x00FF));
			break;
		 case 0x48:	// ST1 w,d
			SETMEMW(GETREGD(REG_1) + GETREGD(GETMEMW(PC++)), GETREGW(opcode & 0x00FF));
			break;
		 case 0x49:	// ST1 d,d
			SETMEMD(GETREGD(REG_1) + GETREGD(GETMEMW(PC++)), GETREGD(opcode & 0x00FF));
			break;
		 case 0x4A:	// ST2 b,d
			SETMEMB(GETREGD(REG_2) + GETREGD(GETMEMW(PC++)), GETREGB(opcode & 0x00FF));
			break;
		 case 0x4B:	// ST2 w,d
			SETMEMW(GETREGD(REG_2) + GETREGD(GETMEMW(PC++)), GETREGW(opcode & 0x00FF));
			break;
		 case 0x4C:	// ST2 d,d
			SETMEMD(GETREGD(REG_2) + GETREGD(GETMEMW(PC++)), GETREGD(opcode & 0x00FF));
			break;
		 case 0x4D:	// STS b,d
			SETMEMB(pKode + GETREGD(REG_4) + GETREGD(GETMEMW(PC++)), GETREGB(opcode & 0x00FF));
			break;
		 case 0x4E:	// STS w,d
			SETMEMW(pKode + GETREGD(REG_4) + GETREGD(GETMEMW(PC++)), GETREGW(opcode & 0x00FF));
			break;
		 case 0x4F:	// STS d,d
			SETMEMD(pKode + GETREGD(REG_4) + GETREGD(GETMEMW(PC++)), GETREGD(opcode & 0x00FF));
			break;
		 case 0x50:	// PU0 b,d
			w = GETMEMW(PC++);
			SETMEMB(pKode + GETREGD(w), GETREGB(opcode & 0x00FF));
			SETREGD(w, GETREGD(w)+1);
			break;
		 case 0x51:	// PU0 w,d
			w = GETMEMW(PC++);
			SETMEMW(pKode + GETREGD(w), GETREGW(opcode & 0x00FF));
			SETREGD(w, GETREGD(w)+2);
			break;
		 case 0x52:	// PU0 d,d
			w = GETMEMW(PC++);
			SETMEMD(pKode + GETREGD(w), GETREGD(opcode & 0x00FF));
			SETREGD(w, GETREGD(w)+4);
			break;
		 case 0x53:	// PO0 b,d
			w = GETMEMW(PC++);
			SETREGD(w, GETREGD(w)-1);
			SETREGB(opcode & 0x00FF, GETMEMB(pKode + GETREGD(w)));
			break;
		 case 0x54:	// PO0 w,d
			w = GETMEMW(PC++);
			SETREGD(w, GETREGD(w)-2);
			SETREGW(opcode & 0x00FF, GETMEMW(pKode + GETREGD(w)));
			break;
		 case 0x55:	// PO0 d,d
			w = GETMEMW(PC++);
			SETREGD(w, GETREGD(w)-4);
			SETREGD(opcode & 0x00FF, GETMEMD(pKode + GETREGD(w)));
			break;
		 case 0x56:	// unused
		 case 0x57:	// unused
			break;
		 case 0x58:	// LD0 b,d,=dval
			w = GETMEMW(PC++);
			SETREGB(opcode & 0x00FF, GETMEMB(pKode + GETREGD(w) + GETMEMD(PC++)));
			PC++;
			break;
		 case 0x59:	// LD0 w,d,=dval
			w = GETMEMW(PC++);
			SETREGW(opcode & 0x00FF, GETMEMW(pKode + GETREGD(w) + GETMEMD(PC++)));
			PC++;
			break;
		 case 0x5A:	// LD0 d,d,=dval
			w = GETMEMW(PC++);
			SETREGD(opcode & 0x00FF, GETMEMD(pKode + GETREGD(w) + GETMEMD(PC++)));
			PC++;
			break;
		 case 0x5B:	// LD1 b,d,=dval
			w = GETMEMW(PC++);
			SETREGB(opcode & 0x00FF, GETMEMB(GETREGD(REG_1) + GETREGD(w) + GETMEMD(PC++)));
			PC++;
			break;
		 case 0x5C:	// LD1 w,d,=dval
			w = GETMEMW(PC++);
			SETREGW(opcode & 0x00FF, GETMEMW(GETREGD(REG_1) + GETREGD(w) + GETMEMD(PC++)));
			PC++;
			break;
		 case 0x5D:	// LD1 d,d,=dval
			w = GETMEMW(PC++);
			SETREGD(opcode & 0x00FF, GETMEMD(GETREGD(REG_1) + GETREGD(w) + GETMEMD(PC++)));
			PC++;
			break;
		 case 0x5E:	// LD2 b,d,=dval
			w = GETMEMW(PC++);
			SETREGB(opcode & 0x00FF, GETMEMB(GETREGD(REG_2) + GETREGD(w) + GETMEMD(PC++)));
			PC++;
			break;
		 case 0x5F:	// LD2 w,d,=dval
			w = GETMEMW(PC++);
			SETREGW(opcode & 0x00FF, GETMEMW(GETREGD(REG_2) + GETREGD(w) + GETMEMD(PC++)));
			PC++;
			break;
		 case 0x60:	// LD2 d,d,=dval
			w = GETMEMW(PC++);
			SETREGD(opcode & 0x00FF, GETMEMD(GETREGD(REG_2) + GETREGD(w) + GETMEMD(PC++)));
			PC++;
			break;
		 case 0x61:	// LDS b,d,=dval
			w = GETMEMW(PC++);
			SETREGB(opcode & 0x00FF, GETMEMB(pKode + GETREGD(REG_4) + GETREGD(w) + GETMEMD(PC++)));
			PC++;
			break;
		 case 0x62:	// LDS w,d,=dval
			w = GETMEMW(PC++);
			SETREGW(opcode & 0x00FF, GETMEMW(pKode + GETREGD(REG_4) + GETREGD(w) + GETMEMD(PC++)));
			PC++;
			break;
		 case 0x63:	// LDS d,d,=dval
			w = GETMEMW(PC++);
			SETREGD(opcode & 0x00FF, GETMEMD(pKode + GETREGD(REG_4) + GETREGD(w) + GETMEMD(PC++)));
			PC++;
			break;
		 case 0x64:	// ST0 b,d,=dval
			w = GETMEMW(PC++);
			SETMEMB(pKode + GETREGD(w) + GETMEMD(PC++), GETREGB(opcode & 0x00FF));
			PC++;
			break;
		 case 0x65:	// ST0 w,d,=dval
			w = GETMEMW(PC++);
			SETMEMW(pKode + GETREGD(w) + GETMEMD(PC++), GETREGW(opcode & 0x00FF));
			PC++;
			break;
		 case 0x66:	// ST0 d,d,=dval
			w = GETMEMW(PC++);
			SETMEMD(pKode + GETREGD(w) + GETMEMD(PC++), GETREGD(opcode & 0x00FF));
			PC++;
			break;
		 case 0x67:	// ST1 b,d,=dval
			w = GETMEMW(PC++);
			SETMEMB(GETREGD(REG_1) + GETREGD(w) + GETMEMD(PC++), GETREGB(opcode & 0x00FF));
			PC++;
			break;
		 case 0x68:	// ST1 w,d,=dval
			w = GETMEMW(PC++);
			SETMEMW(GETREGD(REG_1) + GETREGD(w) + GETMEMD(PC++), GETREGW(opcode & 0x00FF));
			PC++;
			break;
		 case 0x69:	// ST1 d,d,=dval
			w = GETMEMW(PC++);
			SETMEMD(GETREGD(REG_1) + GETREGD(w) + GETMEMD(PC++), GETREGD(opcode & 0x00FF));
			PC++;
			break;
		 case 0x6A:	// ST2 b,d,=dval
			w = GETMEMW(PC++);
			SETMEMB(GETREGD(REG_2) + GETREGD(w) + GETMEMD(PC++), GETREGB(opcode & 0x00FF));
			PC++;
			break;
		 case 0x6B:	// ST2 w,d,=dval
			w = GETMEMW(PC++);
			SETMEMW(GETREGD(REG_2) + GETREGD(w) + GETMEMD(PC++), GETREGW(opcode & 0x00FF));
			PC++;
			break;
		 case 0x6C:	// ST2 d,d,=dval
			w = GETMEMW(PC++);
			SETMEMD(GETREGD(REG_2) + GETREGD(w) + GETMEMD(PC++), GETREGD(opcode & 0x00FF));
			PC++;
			break;
		 case 0x6D:	// STS b,d,=dval
			w = GETMEMW(PC++);
			dw= GETMEMD(PC++);
			PC++;
			SETMEMB(pKode + GETREGD(REG_4) + GETREGD(w) + dw, GETREGB(opcode & 0x00FF));
			break;
		 case 0x6E:	// STS w,d,=dval
			w = GETMEMW(PC++);
			dw= GETMEMD(PC++);
			PC++;
			SETMEMW(pKode + GETREGD(REG_4) + GETREGD(w) + dw, GETREGW(opcode & 0x00FF));
			break;
		 case 0x6F:	// STS d,d,=dval
			w = GETMEMW(PC++);
			dw= GETMEMD(PC++);
			PC++;
			SETMEMD(pKode + GETREGD(REG_4) + GETREGD(w) + dw, GETREGD(opcode & 0x00FF));
			break;
		 case 0x70:	// PU0 b,d,=dval
			w = GETMEMW(PC++);
			dw= GETMEMD(PC++);
			PC++;
			SETMEMB(pKode + GETREGD(w) +dw, GETREGB(opcode & 0x00FF));
			SETREGD(w, GETREGD(w)+1);
			break;
		 case 0x71:	// PU0 w,d,=dval
			w = GETMEMW(PC++);
			dw= GETMEMD(PC++);
			PC++;
			SETMEMW(pKode + GETREGD(w) +dw, GETREGW(opcode & 0x00FF));
			SETREGD(w, GETREGD(w)+2);
			break;
		 case 0x72:	// PU0 d,d,=dval
			w = GETMEMW(PC++);
			dw= GETMEMD(PC++);
			PC++;
			SETMEMD(pKode + GETREGD(w) +dw, GETREGD(opcode & 0x00FF));
			SETREGD(w, GETREGD(w)+4);
			break;
		 case 0x73:	// PO0 b,d,=dval
			w = GETMEMW(PC++);
			dw= GETMEMD(PC++);
			PC++;
			SETREGD(w, GETREGD(w)-1);
			SETREGB(opcode & 0x00FF, GETMEMB(pKode + GETREGD(w) + dw));
			break;
		 case 0x74:	// PO0 w,d,=dval
			w = GETMEMW(PC++);
			dw= GETMEMD(PC++);
			PC++;
			SETREGD(w, GETREGD(w)-2);
			SETREGW(opcode & 0x00FF, GETMEMW(pKode + GETREGD(w) + dw));
			break;
		 case 0x75:	// PO0 d,d,=dval
			w = GETMEMW(PC++);
			dw= GETMEMD(PC++);
			PC++;
			SETREGD(w, GETREGD(w)-4);
			SETREGD(opcode & 0x00FF, GETMEMD(pKode + GETREGD(w) + dw));
			break;
		 case 0x76:	// LSA d,=dval
			SETREGD(opcode & 0x00FF, GETREGD(REG_4) + (int)GETMEMW(PC++));
			break;
		 case 0x77:	// B2I r,r (byte-2-index dst-cnt-reg,src-bit-test-reg
			w = GETMEMW(PC++);
			if( w<0x0080 ) dw = GETREGD(w);
			else if( w<0x00E0 ) dw = (DWORD)GETREGW(w);
			else dw = (DWORD)GETREGB(w);
			if( dw ) {
				result = 0;
				while( !(dw & 1) ) {
					++result;
					dw >>= 1;
				}
			} else result = 0xFFFFFFFF;
			if( dr<0x0080 ) SETREGD(opcode & 0x00FF, result);
			else if( dr<0x00E0 ) SETREGW(opcode & 0x00FF, (WORD)result);
			else SETREGB(opcode & 0x00FF, (BYTE)result);
			break;
		 case 0x78:	// CLR b,b
			dr = opcode & 0x00FF;
			memset(Kregs+dr, 0, GETMEMW(PC++)+1-dr);
			break;
		 case 0x79:	// PUR b,b
			dr = opcode & 0x00FF;
			w = GETMEMW(PC++)+1-dr;
			dw = GETREGD(REG_5);
			memcpy(pKode + dw, Kregs+dr, w);
			SETREGD(REG_5, dw+w);
			break;
		 case 0x7A:	// POR b,b
			dr = opcode & 0x00FF;
			w = GETMEMW(PC++)+1-dr;
			dw = GETREGD(REG_5)-w;
			SETREGD(REG_5, dw);
			memcpy(Kregs+dr, pKode+dw, w);
			break;
		 case 0x7B:	// unused
		 case 0x7C:	// unused
			break;
		 case 0x7D:	// SAF b
			dw = result & 0x00000001;
			if( result==0 ) dw |= 2;
			if( result & 0x80000000 ) dw |= 4;
			if( signedmath ) dw |= 8;
			SETREGB(opcode & 0x00FF, (BYTE)dw);
			break;
		 case 0x7E:	// REF b
			b = GETREGB(opcode & 0x00FF);
			result = ((~b) & 2) | (b & 1);
			if( b & 4 ) result |= 0x80000000;
			signedmath = (b & 8)?1:0;
			break;
		 case 0x7F:	// RET b,=0
			SETREGD(REG_5, GETREGD(REG_5)-4);	// pop return address
			PC = (WORD *)(pKode + GETMEMD(pKode + GETREGD(REG_5)));
			result = (DWORD)SETREGB(opcode & 0x00FF, 0);
			break;
		 case 0x80:	// RET w,=0
			SETREGD(REG_5, GETREGD(REG_5)-4);	// pop return address
			PC = (WORD *)(pKode + GETMEMD(pKode + GETREGD(REG_5)));
			result = (DWORD)SETREGW(opcode & 0x00FF, 0);
			break;
		 case 0x81:	// RET d,=0
			SETREGD(REG_5, GETREGD(REG_5)-4);	// pop return address
			PC = (WORD *)(pKode + GETMEMD(pKode + GETREGD(REG_5)));
			result = (DWORD)SETREGD(opcode & 0x00FF, 0);
			break;
		 case 0x82:	// RET b,=1
			SETREGD(REG_5, GETREGD(REG_5)-4);	// pop return address
			PC = (WORD *)(pKode + GETMEMD(pKode + GETREGD(REG_5)));
			result = (DWORD)SETREGB(opcode & 0x00FF, 1);
			break;
		 case 0x83:	// RET w,=1
			SETREGD(REG_5, GETREGD(REG_5)-4);	// pop return address
			PC = (WORD *)(pKode + GETMEMD(pKode + GETREGD(REG_5)));
			result = (DWORD)SETREGW(opcode & 0x00FF, 1);
			break;
		 case 0x84:	// RET d,=1
			SETREGD(REG_5, GETREGD(REG_5)-4);	// pop return address
			PC = (WORD *)(pKode + GETMEMD(pKode + GETREGD(REG_5)));
			result = (DWORD)SETREGD(opcode & 0x00FF, 1);
			break;
		 case 0x85:	// INC b
			dr = opcode & 0x00FF;
			result = (DWORD)(char)SETREGB(dr, GETREGB(dr) + 1);
			result |= (result<<24);
			break;
		 case 0x86:	// INC w
			dr = opcode & 0x00FF;
			result = (DWORD)(short)SETREGW(dr, GETREGW(dr) + 1);
			result |= (result<<16);
			break;
		 case 0x87:	// INC d
			dr = opcode & 0x00FF;
			result = SETREGD(dr, GETREGD(dr) + 1);
			break;
		 case 0x88:	// DEC b
			dr = opcode & 0x00FF;
			result = (DWORD)(char)SETREGB(dr, GETREGB(dr) - 1);
			result |= (result<<24);
			break;
		 case 0x89:	// DEC w
			dr = opcode & 0x00FF;
			result = (DWORD)(short)SETREGW(dr, GETREGW(dr) - 1);
			result |= (result<<16);
			break;
		 case 0x8A:	// DEC d
			dr = opcode & 0x00FF;
			result = SETREGD(dr, GETREGD(dr) - 1);
			break;
		 case 0x8B:	// NOT b
			dr = opcode & 0x00FF;
			result = (DWORD)(char)SETREGB(dr, ~GETREGB(dr));
			result |= (result<<24);
			break;
		 case 0x8C:	// NOT w
			dr = opcode & 0x00FF;
			result = (DWORD)(short)SETREGW(dr, ~GETREGW(dr));
			result |= (result<<16);
			break;
		 case 0x8D:	// NOT d
			dr = opcode & 0x00FF;
			result = SETREGD(dr, ~GETREGD(dr));
			break;
		 case 0x8E:	// NEG b
			dr = opcode & 0x00FF;
			result = (DWORD)(char)SETREGB(dr, -(char)GETREGB(dr));
			result |= (result<<24);
			break;
		 case 0x8F:	// NEG w
			dr = opcode & 0x00FF;
			result = (DWORD)(short)SETREGW(dr, -(short)GETREGW(dr));
			result |= (result<<16);
			break;
		 case 0x90:	// NEG d
			dr = opcode & 0x00FF;
			result = SETREGD(dr, -(long)GETREGD(dr));
			break;
		 case 0x91:	// PUSH b
			SETMEMB(pKode + GETREGD(REG_5), GETREGB(opcode & 0x00FF));
			SETREGD(REG_5, GETREGD(REG_5)+1);
			break;
		 case 0x92:	// PUSH w
			SETMEMW(pKode + GETREGD(REG_5), GETREGW(opcode & 0x00FF));
			SETREGD(REG_5, GETREGD(REG_5)+2);
			break;
		 case 0x93:	// PUSH d
			SETMEMD(pKode + GETREGD(REG_5), GETREGD(opcode & 0x00FF));
			SETREGD(REG_5, GETREGD(REG_5)+4);
			break;
		 case 0x94:	// POP b
			SETREGD(REG_5, GETREGD(REG_5)-1);
			SETREGB(opcode & 0x00FF, GETMEMB(pKode + GETREGD(REG_5)));
			break;
		 case 0x95:	// POP w
			SETREGD(REG_5, GETREGD(REG_5)-2);
			SETREGW(opcode & 0x00FF, GETMEMW(pKode + GETREGD(REG_5)));
			break;
		 case 0x96:	// POP d
			SETREGD(REG_5, GETREGD(REG_5)-4);
			SETREGD(opcode & 0x00FF, GETMEMD(pKode + GETREGD(REG_5)));
			break;
		 case 0x97:	// unused
			break;
		 case 0x98:	// LD0 b,=dval
			SETREGB(opcode & 0x00FF, GETMEMB(pKode + GETMEMD(PC++)));
			PC++;
			break;
		 case 0x99:	// LD0 w,=dval
			SETREGW(opcode & 0x00FF, GETMEMW(pKode + GETMEMD(PC++)));
			PC++;
			break;
		 case 0x9A:	// LD0 d,=dval
			SETREGD(opcode & 0x00FF, GETMEMD(pKode + GETMEMD(PC++)));
			PC++;
			break;
		 case 0x9B:	// LD1 b,=dval
			SETREGB(opcode & 0x00FF, GETMEMB(GETREGD(REG_1) + GETMEMD(PC++)));
			PC++;
			break;
		 case 0x9C:	// LD1 w,=dval
			SETREGW(opcode & 0x00FF, GETMEMW(GETREGD(REG_1) + GETMEMD(PC++)));
			PC++;
			break;
		 case 0x9D:	// LD1 d,=dval
			SETREGD(opcode & 0x00FF, GETMEMD(GETREGD(REG_1) + GETMEMD(PC++)));
			PC++;
			break;
		 case 0x9E:	// LD2 b,=dval
			SETREGB(opcode & 0x00FF, GETMEMB(GETREGD(REG_2) + GETMEMD(PC++)));
			PC++;
			break;
		 case 0x9F:	// LD2 w,=dval
			SETREGW(opcode & 0x00FF, GETMEMW(GETREGD(REG_2) + GETMEMD(PC++)));
			PC++;
			break;
		 case 0xA0:	// LD2 d,=dval
			SETREGD(opcode & 0x00FF, GETMEMD(GETREGD(REG_2) + GETMEMD(PC++)));
			PC++;
			break;
		 case 0xA1:	// LDS b,=dval
			SETREGB(opcode & 0x00FF, GETMEMB(pKode + GETREGD(REG_4) + (DWORD)((long)((short)GETMEMW(PC++)))));
			break;
		 case 0xA2:	// LDS w,=dval
			SETREGW(opcode & 0x00FF, GETMEMW(pKode + GETREGD(REG_4) + (DWORD)((long)((short)GETMEMW(PC++)))));
			break;
		 case 0xA3:	// LDS d,=dval
			SETREGD(opcode & 0x00FF, GETMEMD(pKode + GETREGD(REG_4) + (DWORD)((long)((short)GETMEMW(PC++)))));
			break;
		 case 0xA4:	// ST0 b,=dval
			SETMEMB(pKode + GETMEMD(PC++), GETREGB(opcode & 0x00FF));
			PC++;
			break;
		 case 0xA5:	// ST0 w,=dval
			SETMEMW(pKode + GETMEMD(PC++), GETREGW(opcode & 0x00FF));
			PC++;
			break;
		 case 0xA6:	// ST0 d,=dval
			SETMEMD(pKode + GETMEMD(PC++), GETREGD(opcode & 0x00FF));
			PC++;
			break;
		 case 0xA7:	// ST1 b,=dval
			SETMEMB(GETREGD(REG_1) + GETMEMD(PC++), GETREGB(opcode & 0x00FF));
			PC++;
			break;
		 case 0xA8:	// ST1 w,=dval
			SETMEMW(GETREGD(REG_1) + GETMEMD(PC++), GETREGW(opcode & 0x00FF));
			PC++;
			break;
		 case 0xA9:	// ST1 d,=dval
			SETMEMD(GETREGD(REG_1) + GETMEMD(PC++), GETREGD(opcode & 0x00FF));
			PC++;
			break;
		 case 0xAA:	// ST2 b,=dval
			SETMEMB(GETREGD(REG_2) + GETMEMD(PC++), GETREGB(opcode & 0x00FF));
			PC++;
			break;
		 case 0xAB:	// ST2 w,=dval
			SETMEMW(GETREGD(REG_2) + GETMEMD(PC++), GETREGW(opcode & 0x00FF));
			PC++;
			break;
		 case 0xAC:	// ST2 d,=dval
			SETMEMD(GETREGD(REG_2) + GETMEMD(PC++), GETREGD(opcode & 0x00FF));
			PC++;
			break;
		 case 0xAD:	// STS b,=dval
			SETMEMB(pKode + GETREGD(REG_4) + (DWORD)((long)((short)GETMEMW(PC++))), GETREGB(opcode & 0x00FF));
			break;
		 case 0xAE:	// STS w,=dval
			SETMEMW(pKode + GETREGD(REG_4) + (DWORD)((long)((short)GETMEMW(PC++))), GETREGW(opcode & 0x00FF));
			break;
		 case 0xAF:	// STS d,=dval
			SETMEMD(pKode + GETREGD(REG_4) + (DWORD)((long)((short)GETMEMW(PC++))), GETREGD(opcode & 0x00FF));
			break;
		 case 0xB0:	// IJMP b,=dval
			dw2 = (DWORD)GETREGB(opcode & 0x00FF);
			goto ijmp;
		 case 0xB1:	// IJMP w,=dval
			dw2 = (DWORD)GETREGW(opcode & 0x00FF);
			goto ijmp;
		 case 0xB2:	// IJMP d,=dval
			dw2 = GETREGD(opcode & 0x00FF);
ijmp:
			pMem = pKode + GETMEMD(PC);	// don't bother to advance PC, we're jumping
			dw = GETMEMD(pMem);	// get # of entries in table
			if( dw2>dw ) dw2 = dw;
			++dw2;	// also skip "number of entries" DWORD
			pMem += 4*dw2;	// skip to the entry (default if past end of table)
			PC = (WORD *)(pKode + GETMEMD(pMem));
			break;
		 case 0xB3:	// KJMP b,=dval
			dw2 = (DWORD)GETREGB(opcode & 0x00FF);
			goto kjmp;
		 case 0xB4:	// KJMP w,=dval
			dw2 = (DWORD)GETREGW(opcode & 0x00FF);
			goto kjmp;
		 case 0xB5:	// KJMP d,=dval
			dw2 = GETREGD(opcode & 0x00FF);
kjmp:
			pMem = pKode + GETMEMD(PC);	// don't bother to advance PC, we're jumping
			dw = GETMEMD(pMem);	// get # of entries in table
			pMem += 4;
			while( dw-- ) {
				if( dw2==GETMEMD(pMem) ) {
					pMem += 4;	// skip value, point at jmp address
					break;
				}
				pMem += 8;	// skip value & jump address
			}
			PC = (WORD *)(pKode + GETMEMD(pMem));
			break;
		 case 0xB6:	// JSR d,=dval
			dw = GETREGD(opcode & 0x00FF) + GETMEMD(PC++);
			PC++;
			SETMEMD(pKode + GETREGD(REG_5), (DWORD)((BYTE *)PC - pKode));	// push return address
			SETREGD(REG_5, GETREGD(REG_5)+4);
			PC = (WORD *)(pKode + dw);
			break;
		 case 0xB7:	// JSI d,=dval
			dw = GETMEMD(pKode + (GETREGD(opcode & 0x00FF) + GETMEMD(PC++)));
			PC++;
			SETMEMD(pKode + GETREGD(REG_5), (DWORD)((BYTE *)PC - pKode));	// push return address
			SETREGD(REG_5, GETREGD(REG_5)+4);
			PC = (WORD *)(pKode + dw);
			break;
		 case 0xB8:	// unused
		 case 0xB9:	// unused
			break;
		 case 0xBA:	// LDI b,=wval
			SETREGB(opcode & 0x00FF, GETMEMW(PC++));
			break;
		 case 0xBB:	// LDI w,=wval
			SETREGW(opcode & 0x00FF, GETMEMW(PC++));
			break;
		 case 0xBC:	// LDI d,=wval
			SETREGD(opcode & 0x00FF, GETMEMD(PC++));
			PC++;
			break;
		 case 0xBD:	// SHL b,=wval
			dr = opcode & 0x00FF;
			result = (DWORD)(char)SETREGB(dr, GETREGB(dr)<<(BYTE)(GETMEMW(PC++) & 0x0007));
			result |= (result<<24);
			break;
		 case 0xBE:	// SHL w,=wval
			dr = opcode & 0x00FF;
			result = SETREGW(dr, GETREGW(dr)<<(BYTE)(GETMEMW(PC++) & 0x000F));
			result |= (result<<16);
			break;
		 case 0xBF:	// SHL d,=wval
			dr = opcode & 0x00FF;
			result = SETREGD(dr, GETREGD(dr)<<(BYTE)(GETMEMW(PC++) & 0x001F));
			break;
		 case 0xC0:	// SHR b,=wval
			dr = opcode & 0x00FF;
			if( signedmath ) result = (DWORD)(char)(SETREGB(dr, (char)GETREGB(dr)>>(BYTE)(GETMEMW(PC++) & 0x0007)));
			else result = (DWORD)(char)(SETREGB(dr, GETREGB(dr)>>(BYTE)(GETMEMW(PC++) & 0x0007)));
			result |= (result<<24);
			break;
		 case 0xC1:	// SHR w,=wval
			dr = opcode & 0x00FF;
			if( signedmath ) result = (DWORD)(short)(SETREGW(dr, (short)(GETREGW(dr))>>(BYTE)(GETMEMW(PC++) & 0x000F)));
			else result = (DWORD)(short)(SETREGW(dr, GETREGW(dr)>>(BYTE)(GETMEMW(PC++) & 0x000F)));
			result |= (result<<16);
			break;
		 case 0xC2:	// SHR d,=wval
			dr = opcode & 0x00FF;
			if( signedmath ) result = SETREGD(dr, (long)GETREGD(dr)>>(BYTE)(GETMEMW(PC++) & 0x001F));
			else result = SETREGD(dr, GETREGD(dr)>>(BYTE)(GETMEMW(PC++) & 0x001F));
			break;
		 case 0xC3:	// ROL b,=wval
			dr = opcode & 0x00FF;
			w = (GETMEMW(PC++) & 0x0007);
			b = GETREGB(dr);
			while( w-- ) {
				if( b & 0x80 ) b = (b << 1) | 1;
				else b <<= 1;
			}
			result = (DWORD)(char)SETREGB(dr, b);
			result |= (result<<24);
			break;
		 case 0xC4:	// ROL w,=wval
			dr = opcode & 0x00FF;
			b = (BYTE)GETMEMW(PC++) & 0x001F;
			w = GETREGW(dr);
			while( b-- ) {
				if( w & 0x8000 ) w = (w << 1) | 1;
				else w <<= 1;
			}
			result = (DWORD)(short)SETREGW(dr, w);
			result |= (result<<16);
			break;
		 case 0xC5:	// ROL d,=wval
			dr = opcode & 0x00FF;
			dw = GETREGD(dr);
			w = GETMEMW(PC++) & 0x1F;
			while( w-- ) {
				dw2 = dw;
				dw <<= 1;
				if( dw2 & 0x80000000 ) dw |= 1;
			}
			result = SETREGD(dr, dw);
			break;
		 case 0xC6:	// ROR b,=wval
			dr = opcode & 0x00FF;
			w = GETMEMW(PC++) & 0x0007;
			b = GETREGB(dr);
			while( w-- ) {
				if( b & 1 ) b = (b >> 1) | 0x80;
				else b >>= 1;
			}
			result = (DWORD)(char)SETREGB(dr, b);
			result |= (result<<24);
			break;
		 case 0xC7:	// ROR w,=wval
			dr = opcode & 0x00FF;
			b = (BYTE)GETMEMW(PC++) & 0x001F;
			w = GETREGW(dr);
			while( b-- ) {
				if( w & 1 ) w = (w >> 1) | 0x8000;
				else w >>= 1;
			}
			result = (DWORD)(short)SETREGW(dr, w);
			result |= (result<<16);
			break;
		 case 0xC8:	// ROR d,=wval
			dr = opcode & 0x00FF;
			dw = GETREGD(dr);
			w = GETMEMW(PC++) & 0x1F;
			while( w-- ) {
				dw2 = dw;
				dw >>= 1;
				if( dw2 & 1 ) dw |= 0x80000000;
			}
			result = SETREGD(dr, dw);
			break;
		 case 0xC9:	// ADD b,=val
			dr = opcode & 0x00FF;
			SETREGB(dr, GETREGB(dr)+(BYTE)GETMEMW(PC++));
			break;
		 case 0xCA:	// ADD w,=val
			dr = opcode & 0x00FF;
			SETREGW(dr, GETREGW(dr)+GETMEMW(PC++));
			break;
		 case 0xCB:	// ADD d,=val
			dr = opcode & 0x00FF;
			SETREGD(dr, GETREGD(dr) + GETMEMD(PC++));
			++PC;	// skip second WORD
			break;
		 case 0xCC:	// SUB b,=val
			dr = opcode & 0x00FF;
			if( signedmath ) result = (DWORD)((long)(((long)((char)GETREGB(dr)))-((long)((char)((BYTE)GETMEMW(PC++))))));
			else result = (DWORD)GETREGB(dr)-(DWORD)GETMEMB(PC++);
			SETREGB(dr, (BYTE)result);
			break;
		 case 0xCD:	// SUB w,=val
			dr = opcode & 0x00FF;
			if( signedmath ) result = (DWORD)((long)(((long)((short)GETREGW(dr)))-((long)((short)(GETMEMW(PC++))))));
			else result = (DWORD)GETREGW(dr)-(DWORD)GETMEMW(PC++);
			SETREGW(dr, (WORD)result);
			break;
		 case 0xCE:	// SUB d,=val
			dr = opcode & 0x00FF;
			if( signedmath ) result = SETREGD(dr, (long)(GETREGD(dr)) - (long)(GETMEMD(PC++)));
			else result = SETREGD(dr, GETREGD(dr) - GETMEMD(PC++));
			++PC;	// skip second WORD
			break;
		 case 0xCF:	// MUL b,=val
			dr = opcode & 0x00FF;
			b = (BYTE)GETMEMW(PC++);
			SETREGB(dr, GETREGB(dr)*b);
			break;
		 case 0xD0:	// MUL w,=val
			dr = opcode & 0x00FF;
			SETREGW(dr, GETREGW(dr)*GETMEMW(PC++));
			break;
		 case 0xD1:	// MUL d,=val
			dr = opcode & 0x00FF;
			SETREGD(dr, GETREGD(dr) * GETMEMD(PC++));
			++PC;	// skip second WORD
			break;
		 case 0xD2:	// DIV b,=val
			dr = opcode & 0x00FF;
			b = (BYTE)GETMEMW(PC++);
			if( signedmath ) SETREGB(dr, (char)(GETREGB(dr))/(char)b);
			else SETREGB(dr, GETREGB(dr)/b);
			break;
		 case 0xD3:	// DIV w,=val
			dr = opcode & 0x00FF;
			w2 = GETMEMW(PC++);
			if( signedmath ) SETREGW(dr, (short)(GETREGW(dr))/(short)w2);
			else SETREGW(dr, GETREGW(dr)/w2);
			break;
		 case 0xD4:	// DIV d,=val
			dr = opcode & 0x00FF;
			if( signedmath ) SETREGD(dr, (long)(GETREGD(dr)) / (long)(GETMEMD(PC++)));
			else SETREGD(dr, GETREGD(dr) / GETMEMD(PC++));
			++PC;	// skip second WORD
			break;
		 case 0xD5:	// MOD b,=val
			dr = opcode & 0x00FF;
			if( signedmath ) SETREGB(dr, (char)(GETREGB(dr))%(char)((BYTE)GETMEMW(PC++)));
			else SETREGB(dr, GETREGB(dr)%((BYTE)GETMEMW(PC++)));
			break;
		 case 0xD6:	// MOD w,=val
			dr = opcode & 0x00FF;
			if( signedmath ) SETREGW(dr, (short)(GETREGW(dr))%(short)((WORD)GETMEMW(PC++)));
			else SETREGW(dr, (WORD)GETREGW(dr)%((WORD)GETMEMW(PC++)));
			break;
		 case 0xD7:	// MOD d,=val
			dr = opcode & 0x00FF;
			if( signedmath ) SETREGD(dr, (long)(GETREGD(dr)) % (long)(GETMEMD(PC++)));
			else SETREGD(dr, GETREGD(dr) % GETMEMD(PC++));
			++PC;	// skip second WORD
			break;
		 case 0xD8:	// OR b,=val
			dr = opcode & 0x00FF;
			SETREGB(dr, GETREGB(dr) | (BYTE)GETMEMW(PC++));
			break;
		 case 0xD9:	// OR w,=val
			dr = opcode & 0x00FF;
			SETREGW(dr, GETREGW(dr) | GETMEMW(PC++));
			break;
		 case 0xDA:	// OR d,=val
			dr = opcode & 0x00FF;
			SETREGD(dr, GETREGD(dr) | GETMEMD(PC++));
			++PC;	// skip second WORD
			break;
		 case 0xDB:	// AND b,=val
			dr = opcode & 0x00FF;
			result = (DWORD)(char)SETREGB(dr, GETREGB(dr) & (BYTE)GETMEMW(PC++));
			result |= (result<<24);
			break;
		 case 0xDC:	// AND w,=val
			dr = opcode & 0x00FF;
			result = (DWORD)(short)SETREGW(dr, GETREGW(dr) & GETMEMW(PC++));
			result |= (result<<16);
			break;
		 case 0xDD:	// AND d,=val
			dr = opcode & 0x00FF;
			result = SETREGD(dr, GETREGD(dr) & GETMEMD(PC++));
			++PC;	// skip second WORD
			break;
		 case 0xDE:	// XOR b,=val
			dr = opcode & 0x00FF;
			result = (DWORD)(char)SETREGB(dr, GETREGB(dr) ^ (BYTE)GETMEMW(PC++));
			result |= (result<<24);
			break;
		 case 0xDF:	// XOR w,=val
			dr = opcode & 0x00FF;
			result = (DWORD)(short)SETREGW(dr, GETREGW(dr) ^ GETMEMW(PC++));
			result |= (result<<16);
			break;
		 case 0xE0:	// XOR d,=val
			dr = opcode & 0x00FF;
			result = SETREGD(dr, GETREGD(dr) ^ GETMEMD(PC++));
			++PC;	// skip second WORD
			break;
		 case 0xE1:	// CMP b,=val
			if( signedmath ) result = (DWORD)((long)(((long)(char)GETREGB(opcode & 0x00FF))-((long)(char)(GETMEMW(PC++)))));
			else result = (DWORD)((DWORD)((long)((short)((char)((BYTE)GETREGB(opcode & 0x00FF)))))-(DWORD)((long)((short)((char)((BYTE)GETMEMW(PC++))))));
			break;
		 case 0xE2:	// CMP w,=val
			if( signedmath ) result = (DWORD)((long)(((long)(short)GETREGW(opcode & 0x00FF))-((long)(short)(GETMEMW(PC++)))));
			else result = (DWORD)GETREGW(opcode & 0x00FF)-(DWORD)GETMEMW(PC++);
			break;
		 case 0xE3:	// CMP d,=val
			if( signedmath ) result = (long)(GETREGD(opcode & 0x00FF)) - (long)(GETMEMD(PC++));
			else result = GETREGD(opcode & 0x00FF) - GETMEMD(PC++);
			++PC;	// skip second WORD
			break;
		 case 0xE4:	// TST b,=val
			result = (DWORD)(char)(GETREGB(opcode & 0x00FF) & (BYTE)GETMEMW(PC++));
			result |= (result<<24);
			break;
		 case 0xE5:	// TST w,=val
			result = (DWORD)(short)(GETREGW(opcode & 0x00FF) & GETMEMW(PC++));
			result |= (result<<16);
			break;
		 case 0xE6:	// TST d,=val
			result = GETREGD(opcode & 0x00FF) & GETMEMD(PC++);
			++PC;	// skip second WORD
			break;
		 case 0xE7:	// JMP =val
			dw = GETMEMD(PC);
			PC = (WORD *)(pKode + dw);
			break;
		 case 0xE8:	// JE  +offset (0 to 255 WORDS from CP)
			if( !result ) PC += (DWORD)(opcode & 0x01FF);
			break;
		 case 0xE9:	// JE  -offset (-256 to -1 WORDS from CP)
			if( !result ) PC += (DWORD)(opcode & 0x01FF) | 0xFFFFFE00;
			break;
		 case 0xEA:	// JNE +offset
			if( result ) PC += (DWORD)(opcode & 0x01FF);
			break;
		 case 0xEB:	// JNE -offset
			if( result ) PC += (DWORD)(opcode & 0x01FF) | 0xFFFFFE00;
			break;
		 case 0xEC:	// JL  +offset
			if( result & 0x80000000 ) PC += (DWORD)(opcode & 0x01FF);
			break;
		 case 0xED:	// JL  -offset
			if( result & 0x80000000 ) PC += (DWORD)(opcode & 0x01FF) | 0xFFFFFE00;
			break;
		 case 0xEE:	// JLE +offset
			if( !result || (result & 0x80000000) ) PC += (DWORD)(opcode & 0x01FF);
			break;
		 case 0xEF:	// JLE -offset
			if( !result || (result & 0x80000000) ) PC += (DWORD)(opcode & 0x01FF) | 0xFFFFFE00;
			break;
		 case 0xF0:	// JG  +offset
			if( result && !(result & 0x80000000) ) PC += (DWORD)(opcode & 0x01FF);
			break;
		 case 0xF1:	// JG  -offset
			if( result && !(result & 0x80000000) ) PC += (DWORD)(opcode & 0x01FF) | 0xFFFFFE00;
			break;
		 case 0xF2:	// JGE +offset
			if( !(result & 0x80000000) ) PC += (DWORD)(opcode & 0x01FF);
			break;
		 case 0xF3:	// JGE -offset
			if( !(result & 0x80000000) ) PC += (DWORD)(opcode & 0x01FF) | 0xFFFFFE00;
			break;
		 case 0xF4:	// JEV +offset
			if( !(result & 1) ) PC += (DWORD)(opcode & 0x01FF);
			break;
		 case 0xF5:	// JEV -offset
			if( !(result & 1) ) PC += (DWORD)(opcode & 0x01FF) | 0xFFFFFE00;
			break;
		 case 0xF6:	// JOD +offset
			if( result & 1 ) PC += (DWORD)(opcode & 0x01FF);
			break;
		 case 0xF7:	// JOD -offset
			if( result & 1 ) PC += (DWORD)(opcode & 0x01FF) | 0xFFFFFE00;
			break;
		 case 0xF8:	// JMP +offset
			PC += (DWORD)(opcode & 0x01FF);
			break;
		 case 0xF9:	// JMP -offset
			PC += (DWORD)(opcode & 0x01FF) | 0xFFFFFE00;
			break;
		 case 0xFA:	// SysFunc
			SysFunc(Kregs, opcode & 0x00FF);
			break;
		 case 0xFB:	// JSR =dval
			dw = GETMEMD(PC++);
			PC++;
			SETMEMD(pKode + GETREGD(REG_5), (DWORD)((BYTE *)PC - pKode));	// push return address
			SETREGD(REG_5, GETREGD(REG_5)+4);
			PC = (WORD *)(pKode + dw);
			break;
		 case 0xFC:	// JSI =dval
			dw = GETMEMD(pKode + GETMEMD(PC++));
			++PC;
			SETMEMD(pKode + GETREGD(REG_5), (DWORD)((BYTE *)PC - pKode));	// push return address
			SETREGD(REG_5, GETREGD(REG_5)+4);
			PC = (WORD *)(pKode + dw);
			break;
		 case 0xFD:	// RET
			SETREGD(REG_5, GETREGD(REG_5)-4);	// pop return address
			PC = (WORD *)(pKode + GETMEMD(pKode + GETREGD(REG_5)));
			break;
		 case 0xFE:	// misc
			switch( opcode & 0x00FF ) {
			 case 0x0000:	// BRK
				break;
			 case 0x0001:	// NOP
				break;
			 case 0x0002:	// TERM
				killpgm |= 1;	// set flag we need to kill the program
// NOTE: case TERM falls through into case HALT
			 case 0x0003:	// HALT
				--PC;	// move PC back to TERM instruction just for safety
				SETREGD(REG_3, (DWORD)((BYTE *)PC - pKode));	// update PC in Kregs
				goto stopint;
			 case 0x0004:	// AFR
				// pu0 R4,R5
				SETMEMD(pKode + GETREGD(REG_5), GETREGD(REG_4));
				SETREGD(REG_5, GETREGD(REG_5)+4);
				// ldr R4,R5
				SETREGD(REG_4, GETREGD(REG_5));
				// add R5,=*PC++
				SETREGD(REG_5, GETREGD(REG_5) + GETMEMW(PC++));
				break;
			 case 0x0005:	// DFR
				// ldr R5,R4
				SETREGD(REG_5, GETREGD(REG_4)-4);	// minus the "pop" that's coming up
				// po0 R4,R5
				SETREGD(REG_4, GETMEMD(pKode+GETREGD(REG_5)));
				break;
			 case 0x0006:	// USM
				signedmath = 0;
				break;
			 case 0x0007:	// SM
				signedmath = 1;
				break;
			 case 0x0010:	// LDCM b,b
				w = GETMEMW(PC++);
				SETREGB(w & 0x00FF, GETREGB(w & 0x00FF)-1);
				SETREGB((w>>8) & 0x00FF, GETREGB(GETREGB(w & 0x00FF)));
				break;
			 case 0x0011:	// LDCM w,b
				w = GETMEMW(PC++);
				SETREGB(w & 0x00FF, GETREGB(w & 0x00FF)-2);
				SETREGW((w>>8) & 0x00FF, GETREGW(GETREGB(w & 0x00FF)));
				break;
			 case 0x0012:	// LDCM d,b
				w = GETMEMW(PC++);
				SETREGB(w & 0x00FF, GETREGB(w & 0x00FF)-4);
				SETREGD((w>>8) & 0x00FF, GETREGD(GETREGB(w & 0x00FF)));
				break;
			 case 0x0013:	// LDC b,b
				w = GETMEMW(PC++);
				SETREGB((w>>8) & 0x00FF, GETREGB(GETREGB(w & 0x00FF)));
				break;
			 case 0x0014:	// LDC w,b
				w = GETMEMW(PC++);
				SETREGW((w>>8) & 0x00FF, GETREGW(GETREGB(w & 0x00FF)));
				break;
			 case 0x0015:	// LDC d,b
				w = GETMEMW(PC++);
				SETREGD((w>>8) & 0x00FF, GETREGD(GETREGB(w & 0x00FF)));
				break;
			 case 0x0016:	// LDCP b,b
				w = GETMEMW(PC++);
				SETREGB((w>>8) & 0x00FF, GETREGB(GETREGB(w & 0x00FF)));
				SETREGB(w & 0x00FF, GETREGB(w & 0x00FF)+1);
				break;
			 case 0x0017:	// LDCP w,b
				w = GETMEMW(PC++);
				SETREGW((w>>8) & 0x00FF, GETREGW(GETREGB(w & 0x00FF)));
				SETREGB(w & 0x00FF, GETREGB(w & 0x00FF)+2);
				break;
			 case 0x0018:	// LDCP d,b
				w = GETMEMW(PC++);
				SETREGD((w>>8) & 0x00FF, GETREGD(GETREGB(w & 0x00FF)));
				SETREGB(w & 0x00FF, GETREGB(w & 0x00FF)+4);
				break;
			 case 0x0019:	// STCM b,b
				w = GETMEMW(PC++);
				SETREGB(w & 0x00FF, GETREGB(w & 0x00FF)-1);
				SETREGB(GETREGB(w & 0x00FF), GETREGB((w>>8) & 0x00FF));
				break;
			 case 0x001A:	// STCM w,b
				w = GETMEMW(PC++);
				SETREGB(w & 0x00FF, GETREGB(w & 0x00FF)-2);
				SETREGW(GETREGB(w & 0x00FF), GETREGW((w>>8) & 0x00FF));
				break;
			 case 0x001B:	// STCM d,b
				w = GETMEMW(PC++);
				SETREGB(w & 0x00FF, GETREGB(w & 0x00FF)-4);
				SETREGD(GETREGB(w & 0x00FF), GETREGD((w>>8) & 0x00FF));
				break;
			 case 0x001C:	// STC b,b
				w = GETMEMW(PC++);
				SETREGB(GETREGB(w & 0x00FF), GETREGB((w>>8) & 0x00FF));
				break;
			 case 0x001D:	// STC w,b
				w = GETMEMW(PC++);
				SETREGW(GETREGB(w & 0x00FF), GETREGW((w>>8) & 0x00FF));
				break;
			 case 0x001E:	// STC d,b
				w = GETMEMW(PC++);
				SETREGD(GETREGB(w & 0x00FF), GETREGD((w>>8) & 0x00FF));
				break;
			 case 0x001F:	// STCP b,b
				w = GETMEMW(PC++);
				SETREGB(GETREGB(w & 0x00FF), GETREGB((w>>8) & 0x00FF));
				SETREGB(w & 0x00FF, GETREGB(w & 0x00FF)+1);
				break;
			 case 0x0020:	// STCP w,b
				w = GETMEMW(PC++);
				SETREGW(GETREGB(w & 0x00FF), GETREGW((w>>8) & 0x00FF));
				SETREGB(w & 0x00FF, GETREGB(w & 0x00FF)+2);
				break;
			 case 0x0021:	// STCP d,b
				w = GETMEMW(PC++);
				SETREGD(GETREGB(w & 0x00FF), GETREGD((w>>8) & 0x00FF));
				SETREGB(w & 0x00FF, GETREGB(w & 0x00FF)+4);
				break;
			}
			break;
		 case 0xFF:	// unused
			break;
		}
		SETREGD(REG_3, (DWORD)((BYTE *)PC - pKode));	// update PC in Kregs
	}
stopint:
	return GETREGW(REG_39);
}

//*******************************************************
//* This routine is called to invoke the .KE "event"
//* entry point, with:
//* R32 = cmd (offset into pKode...see KHDR_ #defines in pmach.h)
//*	R15 = arg1
//*	R14 = arg2
//*	R13 = arg3
//*
WORD KEevent(WORD cmd, DWORD arg1, DWORD arg2, DWORD arg3) {

	BYTE	*Kregs;
	WORD	retval;

	if( gateEvent ) return 0;	// don't allow re-entrancy (as in SysSetWindowPos)
	gateEvent = 1;

// initialize CPU "system" regs (membase, PC, stack ptrs) and event regs
// WARN: PTR4
	Kregs = mainRegs;	// SETREGx macros use "Kregs", which is what ALL of ExecKE uses.

	SETREGD(REG_0, mainKode);	// init code block base register
	SETREGD(REG_3, GETMEMD(mainKode+cmd));	// init PC
	if( cmd==KHDR_BEGIN_STARTUP ) {	// only init stack when starting up
		SETREGD(REG_4, StackBase);	// init stack BASE frame ptr
		SETREGD(REG_5, StackBase);	// init stack top ptr
	}
	SETREGW(REG_32, cmd);		// set cmd
	SETREGD(REG_15, arg1);		//    and args into CPU regs for Kode
	SETREGD(REG_14, arg2);
	SETREGD(REG_13, arg3);
// call K interpreter
	retval = ExecKE(mainRegs, mainKode);
	gateEvent = 0;
// if Kode did a TERM, kill the program, but ONLY do it ONCE!
// (The Kode may call TERM more than once due to "shut down" Kode calls)
	if( killpgm==1 ) {
		killpgm |= 2;
		KillProgram();
	}
	return retval;
}

